@@ -1,148 +1,242 @@
-# $Id: Exercise.pm 266 2009-06-04 13:48:25Z lembark $
 #######################################################################
 # housekeeping
 #######################################################################
 
-package Object::Exercise;
+package Object::Exercise    v3.0.1;
+use v5.20;
 
-require 5.6.2; # I'm running 5.8.8; hopefully this is reasonable...
+use Symbol qw( qualify qualify_to_ref );
 
-use strict;
-
-use Symbol qw( qualify_to_ref );
-
-use Object::Exercise::Common qw( log_message continue verbose );
+use Object::Exercise::Execute;
 
 ########################################################################
 # package variables
 ########################################################################
 
-our $VERSION = 1.02;
-
 ########################################################################
 # subroutines
 ########################################################################
 
 sub import
 {
-    my $package = __PACKAGE__;
+    state $init     = \&Object::Exercise::Execute::import_flags;
+    state $handler  = \&Object::Exercise::Execute::exercise;
+
     my $caller  = caller;
 
-    # discard the class argument.
+    # ignore the current package
 
-    shift if $_[0] eq $package;
+    shift;
 
-    # arguments:
-    # -k turns off fatal on error.
-    # -v turns on  verbose.
-    # -b uses benchmark instead of execution handler.
-    # -e uses execution handler (default).
-    # -n specifies the installed name (vs. 'execution' or 'benchmark'.
-    # -p turns off plannng in the test loop.
+    my $export  = &$init;
 
-    my %exportz = ();
+    # install the handler as scalar ref not a sub ref so that
+    # they get $exercise as a subref.
 
-    while( @_ )
-    {
-        my $arg = shift;
-
-        if( $arg =~ /^-k/ )
-        {
-            $continue = 1;
-        }
-        elsif( $arg =~ /^-v/ )
-        {
-            $verbose = 1;
-        }
-        elsif( $arg =~ /^-b/ )
-        {
-            my $name = ( index $_[0], '-' ) ? 'benchmark' : shift ;
-
-            $exportz{ benchmark } = $name;
-        }
-        elsif( $arg =~ /^-e/ )
-        {
-            my $name = ( index $_[0], '-' ) ? 'execute' : shift ;
-
-            $exportz{ execute } = $name;
-        }
-        else
-        {
-            die "Bogus $package: unknown switch '$arg'";
-        }
-    }
-
-    %exportz = qw( execute exercise ) unless %exportz;
-
-    while( my($src,$dst) = each %exportz )
-    {
-        $log_message->( "$package installing '$src' into '$caller' as '$dst'" )
-        if $verbose;
-
-        my $module  = $package . '::' . ucfirst $src;
-
-        my $handler = eval "require $module";
-
-        my $ref     = qualify_to_ref $dst, $caller;
-
-        *$ref       = \$handler;
-    }
+    *{ qualify_to_ref $export => $caller } = \$handler;
 
     return
 }
 
 # keep require happy
-
 1
-
 __END__
 
 =head1 NAME
 
-Object::Exercise - Generic execution & benchmark harness for method calls.
+Object::Exercise - Generic execution & benchmark harness
+for method calls.
 
 =head1 SYNOPSIS
 
-  use Object::Exercise;
+    use Object::Exercise;
 
-  my @operationz =
-  (
-    [
-      [ method arg arg arg ... ]          # method and arguments
-      [ 1 ],                              # expected value
-    ],
+    # at this point $exercise is defined as a 
+    # coderef to execute the tests.
 
-    [
-      qw( method arg arg arg )            # just check for $@
-    ],
+    my @test_opz =
+    (
+        [
+            # no arrayref: call $obj->$method( @argz )
+            # check for $@ with message of method and args.
 
-    [
-      [ qw( method expected to fail ) ]   # continue on failure
-      [],
-    ],
+            qw( method arg arg arg )
+        ],
 
-    [
-      [ $coderef, @argz ],                # $obj->$coderef( @argz )
-      [ ( 1 .. 10 )     ],                # expected value
-      'Coderef returns list'              # hardwired message
-    ]
-  );
+        [
+            # two arrayrefs:
+            # [ method and args ], 
+            # [ compare values ]
 
-  # You can push the operations through an class:
+            [ method arg arg arg ... ]
+            [ 1 ],
+        ],
 
-  $exercise->( 'YourClass', @test_opz );    # YourClass->method( @argz )
+        [
+            # if a method is expected to fail, store
+            # a literal 'undef' as the second element.
+            
+            [ qw( method expected to fail ) ]
+            undef,
+        ],
 
-  # or an object:
+        [
+            # use a pre-defined message rather than
+            # join the method and args.
 
-  my $object = YourClass->new( @whatever );
+            [ method => @argz ],
+            [ ( 1 .. 10 )     ],
+            'Coderef returns list'
+        ],
 
-  $exercise->( $object, @test_opz );        # $object->method( @argz )
+        [
+            # use a coderef to dispatch the call rather
+            # than a method name.
+            #
+            # storing a literal undef as an array value
+            # checks for an undef -- vs. ignoring the 
+            # return value in the example above.
 
+            [ $subref, @argz ],
+            [ undef ],
+            'Coderef returns list'
+        ],
+
+        [
+            # compare the method return via regex
+            # effectively:
+            # $obj->Foobar( @argz ) =~ qr/Foo.*Bar/
+
+            [ method => @argz ],
+            qr/Foo.*Bar/,
+            'method returns text of Foo ... Bar'
+        ],
+
+        [
+            # compare the return via user-supplied 
+            # filter (instead of cmp_deeply).
+
+            [ method => @argz ],
+            $subref
+        ],
+
+        # turn on verbosity
+
+        q{verbose}, 
+
+        # subref is called to validate the result when static
+        # values get messy (e.g., database testing or when the
+        # return contains a timestamp or using a closure 
+        # computed when the arguments are constructed).
+
+        [
+            [ blah_blah => ( 'a' .. 'z' ) ],
+            \&validate_blah_blah,
+            'Validate blah_blah'
+        ],
+
+
+        # turn off verbosity
+
+        q{noverbose}
+
+        # turn on/off a breakpoint prior to execution or if the
+        # next step returns an error.
+
+        ...
+
+        # set a breakpoint via $DB::single = 1 prior to calling
+        # $object->$method for all following tests.
+
+        q{break},
+
+        [
+            # for this one test: 
+            # set a breakpoint after calling $object->$method 
+            # iff $@ is true via $DB::single = 1 if $error;
+            # also make this one test verbose.
+
+            debug   =>
+            verbose =>
+
+            [ frobnicate => @frob_argz          ],
+            [ 'eeny', 'meene', 'minie', 'moe'   ],
+            [ "Frobnicate returns gibberish"    ]
+        ],
+        [
+            # YAML::XS has problems (a.k.a. segfaults) 
+            # handling stored regexen. simplest workaround
+            # is setting a flag to treat non-ref expect values
+            # as bare regexen and feed them through qr{}x.
+
+            [
+                regex =>
+                [
+                    'get_title'
+                ],
+                YAPC =>
+                "Title includes 'YAPC'"
+            ],
+        ],
+    );
+
+    # stack tests: 
+    # "noplan" skips calling "plan" with number of tests.
+    # "nofinish" skips calling "done_testing" after current plan.
+    #
+    # this leaves it up to the caller to call "done_testing"
+    # or call exercise with "finish" set to true.
+
+    $obj->$exercise ( qw( nofinish ) );
+
+    $obj->$exercise( @plan1 );
+    $obj->$exercise( @plan2 );
+    $obj->$exercise( @plan3 );
+
+    $obj->$exercise ( 'finish' );   # call's done testing.
+
+    # You can push the operations through as
+    # class or object methods.
+
+    YourClass->$exercise( @test_opz );
+
+    my $object = YourClass->new( @whatever );
+
+    $object->$exercise( @test_opz );
+
+    # or just
+
+    YourClass->new( @blah )->$exercise( @test_opz );
+
+    # sometimes it is simpler to pass the tests in as an arrayref
+    # or as flat YAML text. A single arrayref on the stack will be
+    # expanded in-place to get the tests, a single non-ref scalar
+    # will be treated as YAML.
+    #
+    # this makes it easy to store an arrayref as text or YAML in
+    # the test body:
+
+    my $plan    = do { local $/; eval <DATA> };
+    $obj->$exercise( $plan );
+
+    __DATA__
+    # or this could be a perly arrayref in Dumper format.
+
+    ---
+    - noplan
+    - - verbose
+      - - prepare_test
+        - - get
+          - http://www.google.com
+        - - set_implicit_wait_timeout
+          - 50000
+    ...
+ 
 
 =head1 DESCRIPTION
 
 This package exports a single subroutine , C<$exercise>, which
-functions as an OO execution loop (see '-n' for changing the 
+functions as an OO execution loop (see 'export' for changing the
 installed name).
 
 C<$execute> is a subroutine reference that takes an object
@@ -166,41 +260,60 @@ and its arguments.
 
 =over 4
 
-=item -e [yourname]
+=item install=<yourname> (default "execute")
 
-Default is to install '$execute' to run operations,
-testing for $@, and comparing return values with cmp_deeply.
+Default is to install '$execute', any alternate variable
+name can be selected.
 
-The optional 'yourname' can provide an alternate name
-to '$execute'.
+=item benchmark (default 0)
 
-=item -b [yourname]
+This writes out the wallclock time to four decimal places
+(from Time::Hires) for each entry executed.
 
-Alternative is to intall '$benchmark' with counts the
-operations and errors (via $@), and reports the total
-elapsed time, operations, and errors.
+This can be set inline via [no]benchmark:
 
-The optional 'yourname' can provide an alternate name
-to '$benchmark'.
+    [ ... ],
 
-=item -v 
+    # showing elapsed time for only one test:
 
-Turn on verbose reporting of results in execution mode
-and report the symbols exported, largely equivalent to
-adding 'verbose' before any arrayrefs.
+    'benchmark',
 
-=item -k 
+    [ ... ],
 
-Assume failures are expected and ignore them for logging
-and breakpoints. Equivalent to adding 'continue' before 
-any arrayrefs (i.e., as with "make -k").
+    'nobenchmark',
 
-=item Notes
+=item verbose (default 0)
 
-If neither -e nor -b is used then the default is to supply
-'$execute'.
+Turn on additinal progress messages, printing them
+to stderr so that they show up even if prove is not
+running in verbose mode. 
 
-If both -e and -b are used then both will be exported.
+This can be switched innline via "verbose" or "noverbose".
+
+=item continue (default 1)
+
+Continue execution after failures. Turning this off
+aborts after the first execption, which may interfere
+with test counts (see also "plan", next).
+
+The main use of this is setting "nocontinue" in cases
+where some initial setup steps failing will force the
+remaining steps to fail. 
+
+=item plan (default 1)
+
+Normally an initial count of the tests is used generate
+a test plan. If this is specified then no plan is used.
+This can be helpful when used with "nocontinue".
+
+=item debug (default 0)
+
+This sets a breakpoint if an execption is raised calling
+the method. The line following this breakpoint is:
+
+    $obj->$method( @$argz );
+
+which can be used to immediately check why a call failed.
 
 =back
 
@@ -303,14 +416,16 @@ use something like:
 
   [
     [ $method => @argz ],
-    [ # outer arrayref stored return value
+    [
+        # outer arrayref stored return value
+        # inner value is the hashref expected value.
 
-      { # return value is itself an hashref }
+      { ...  }
     ],
   ],
 
-The default C<ok> message is formed by joining the 
-method and arguments on whitespace. This can lead to 
+The default C<ok> message is formed by joining the
+method and arguments on whitespace. This can lead to
 prove issuing lines like:
 
   ok save foobar HASH(0x123456) (999)
@@ -339,7 +454,12 @@ These tests look like:
   [
     [ qw( method designed to fail ) ]
     '',
-  ]
+  ],
+
+  [
+    [ qw( method designed to fail ) ]
+    undef,
+  ],
 
 This will give a message like:
 
@@ -421,13 +541,13 @@ For example:
   perl -d harness_code.t;
 
 will stop execution at the first failed operation, allowing
-a single C<s> to step into the C<$obj->$method( @argz )> call.
+a single C<s> to step into the C<< $obj->$method( @argz ) >> call.
 
 =head2 Harness Directives
 
 There are times when you want to control the execution
 or harness arguments as it is running. The directives
-are processed by the harness itself. These can set a 
+are processed by the harness itself. These can set a
 breakpoint prior to calling one of the methods, adjust
 the verbosity, set the continue switch, or set an object
 value.
@@ -437,7 +557,7 @@ value.
 Sometimes it's reassuring to know progress is being
 make (or it's helpful to keep a log of what happend).
 
-Non-referent entries in the test list that aren't 
+Non-referent entries in the test list that aren't
 recognized are simply printed:
 
     ...
@@ -457,13 +577,11 @@ will sandwich a method call between two messages.
 
 It is sometimes helpful to stop the execution of code
 before it fails in order to examine its execution before
-the failure. The "debug" directive will set the breakpoint
+the failure. The "break" directive will set the breakpoint
 before the first method call. This will leave you at
 $obj->$method( @argz ) (see below under DEBUGGING).
 
-The 'debug' directive can be aliased as 'break'.
-
-For example, this will run up to the point where 
+For example, this will run up to the point where
 "frobnicate" is about to be called and then stop:
 
   [
@@ -479,26 +597,12 @@ For example, this will run up to the point where
     'nobreak',
   ],
 
-Until "nodebug" or "nobreak" is used after this, all calls
-will hit the breakpoint. Using an expect value of '' turns
-on continue mode for one operation (e.g., for testing 
-proper handling of failure cases):
+This leaves the perl debugger at the line:
 
-    [
-        [ qw( this_fails ) ],
-        ''
-    ],
+    my $result = [ $obj->$method( @$argz ) ];
 
-is equivalent to:
-
-    'continue',
-
-    [
-        [ qw( this_fails        ) ],
-        [ qw( expect failure    ) ],
-    ],
-
-    'nocontinue',
+Until "nobreak" is used all calls will hit the 
+breakpoint.
 
 =head3 Turning off comparison breakpoints.
 
@@ -554,15 +658,17 @@ Re-run a failed operation:
     DB<<2>> s
 
 At this point you will be at the first line of $method
-(given sub or coderef location). The failure message 
+(given sub or coderef location). The failure message
 will show up for $@ set after calling the method or
 if cmp_deeply finds a discrepency in the result.
 
 
 =head1 EXAMPLES
 
+Aside: Yes I know we need more of these...
+
     my $field2 = 'field_x';
-  
+
 
     my @opz =
     (
@@ -588,7 +694,7 @@ if cmp_deeply finds a discrepency in the result.
         ],
 
         [
-            [ write =>  ( label => $field2, '' )    ],  # invalid argument: 
+            [ write =>  ( label => $field2, '' )    ],  # invalid argument:
             '',                                         # expect failure
         ],
 
@@ -608,6 +714,7 @@ Steven Lembark <lembark@wrkhors.com>
 
 =head1 COPYRIGHT
 
-Copyright (C) 2007 Steven Lembark.
-This code is released under the same terms as Perl-5.8.
+Copyright (C) 2007-2015 Steven Lembark.
+This code is released under the same terms as Perl-5.22 or any
+later version of Perl.
 
