@@ -2,10 +2,9 @@
 # housekeeping
 #######################################################################
 
-package Object::Exercise::Execute;
+package Object::Exercise::Execute   v3.0.1;
 use v5.20;
 no warnings;
-use experimental    qw( autoderef );
 
 use Carp;
 
@@ -22,9 +21,6 @@ use Test::More;
 
 our @CARP_NOT   = ( __PACKAGE__ );
 
-our $VERSION    = '3.00';
-$VERSION        = eval "$VERSION";
-
 my @defaultz =
 qw
 (
@@ -189,17 +185,16 @@ my $test_result
             # this may pass if expect is [ undef ] but will
             # report the error text either way.
 
-            ! $type
-            and
-            $globalz{ regex }
-            and
-            do
+            if
+            (
+                ! $type
+                and
+                $globalz{ regex }
+            )
             {
-                $expect = qr{$expect};
-                $type   = 'REGEXP';
-            };
-
-            if( 'REGEXP' eq $type )
+                like    "@$found", qr{$expect}x, $message;
+            }
+            elsif( 'REGEXP' eq $type )
             {
                 like        "@$found", $expect, $message; 
             }
@@ -261,7 +256,7 @@ my $process
         return
     }
 
-    local @globalz{ keys $localz } = values $localz
+    local @globalz{ keys %$localz } = values %$localz
     if $localz;
 
     # if there is no expect value then skip the ok check and just
@@ -399,8 +394,6 @@ sub prepare_tests
 
 sub exercise
 {
-$DB::single = 1;
-
     my $obj     = shift;
 
     # test seqeunce is left on the stack.
